def function_to_text(serialized_function_proto: bytes) -> str: ...
def graph_to_text(serialized_graph_proto: bytes) -> str: ...
def model_to_text(serialized_model_proto: bytes) -> str: ...
def node_to_text(serialized_node_proto: bytes) -> str: ...
